/*
	5TQ->850ϊ
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include "hex.h"
#include "fontmisaki.h"
#include "fontk6x8.h"
#include "font4x8.h"

#define FALSE	0
#define TRUE	1

#ifndef O_BINARY
#	define O_BINARY	0
#	define FOPEN_TEXT	""
#else
#	define FOPEN_TEXT	"t"
#endif

/* R[h\ */
struct CodeTable {
	unsigned char ch[2];	/* (SJIS) */
	short n;	/* o */
};

/*
	SJIS̐擪?
*/
static int isKanji(const unsigned char c)
{
	return
	(0x81 <= c && c <= 0x9f) ||
	(0xe0 <= c && c <= 0xfc);
}

/*
	5TQt@C̕𓾂
*/
static int get5tqText(unsigned char *out, const unsigned char *in, int size, int crypt)
{
	unsigned char *p;
	const unsigned char *q;

	for(p = out, q = in; q < in + size; p++, q++) {
		if(crypt)
			*p = *q ^ 0x80;
		else
			*p = *q;
	}

	for(p = out; p < out + size && *p != (crypt ? (unsigned char )0xa0: 0x20); p += isKanji(*p) ? 2: 1)
		;
	*p = 0;

	return (int )(p - out);
}

/*
	5TQt@C̃wb_ǂ
*/
static int read5tqHead(int fd, unsigned char *category, int *n)
{
	int size;
	unsigned char buf[256];

	if((size = read(fd, buf, sizeof(buf))) <= 0)
		return size;

	get5tqText(category, buf, 16, FALSE);
	*n = (int )buf[19] * 0x100 + (int )buf[18];
	return size;
}

/*
	5TQt@C̃f[^ǂݍ
*/
static int read5tqData(int fd, unsigned char *q, unsigned char *a1, unsigned char *a2, unsigned char *a3, unsigned char *a4, unsigned char *a5)
{
	int size;
	unsigned char buf[256];

	if((size = read(fd, buf, sizeof(buf))) <= 0)
		return size;

	get5tqText(q, buf, 116, TRUE);
	get5tqText(a1, buf + 116 + 28 * 0, 28, TRUE);
	get5tqText(a2, buf + 116 + 28 * 1, 28, TRUE);
	get5tqText(a3, buf + 116 + 28 * 2, 28, TRUE);
	get5tqText(a4, buf + 116 + 28 * 3, 28, TRUE);
	get5tqText(a5, buf + 116 + 28 * 4, 28, TRUE);
	return size;
}

/*
	u
*/
static int replaceChar(unsigned char *out, const unsigned char *in)
{
	const struct {
		unsigned char *src;
		unsigned char *dst;
	} conv[] = {
		{ (unsigned char *)"[", (unsigned char *)"|" },
		{ (unsigned char *)"!", (unsigned char *)"I" },
		{ (unsigned char *)"\"", (unsigned char *)"h" },
		{ (unsigned char *)"#", (unsigned char *)"" },
		{ (unsigned char *)"$", (unsigned char *)"" },
		{ (unsigned char *)"%", (unsigned char *)"" },
		{ (unsigned char *)"&", (unsigned char *)"" },
		{ (unsigned char *)"'", (unsigned char *)"f" },
		{ (unsigned char *)"(", (unsigned char *)"i" },
		{ (unsigned char *)")", (unsigned char *)"j" },
		{ (unsigned char *)"*", (unsigned char *)"" },
		{ (unsigned char *)"+", (unsigned char *)"{" },
		{ (unsigned char *)",", (unsigned char *)"C" },
		{ (unsigned char *)"-", (unsigned char *)"|" },
		{ (unsigned char *)".", (unsigned char *)"D" },
		{ (unsigned char *)"/", (unsigned char *)"^" },
		{ (unsigned char *)"0", (unsigned char *)"O" },
		{ (unsigned char *)"1", (unsigned char *)"P" },
		{ (unsigned char *)"2", (unsigned char *)"Q" },
		{ (unsigned char *)"3", (unsigned char *)"R" },
		{ (unsigned char *)"4", (unsigned char *)"S" },
		{ (unsigned char *)"5", (unsigned char *)"T" },
		{ (unsigned char *)"6", (unsigned char *)"U" },
		{ (unsigned char *)"7", (unsigned char *)"V" },
		{ (unsigned char *)"8", (unsigned char *)"W" },
		{ (unsigned char *)"9", (unsigned char *)"X" },
		{ (unsigned char *)":", (unsigned char *)"F" },
		{ (unsigned char *)";", (unsigned char *)"G" },
		{ (unsigned char *)"<", (unsigned char *)"" },
		{ (unsigned char *)"=", (unsigned char *)"" },
		{ (unsigned char *)">", (unsigned char *)"" },
		{ (unsigned char *)"?", (unsigned char *)"H" },
		{ (unsigned char *)"@", (unsigned char *)"" },
		{ (unsigned char *)"A", (unsigned char *)"`" },
		{ (unsigned char *)"B", (unsigned char *)"a" },
		{ (unsigned char *)"C", (unsigned char *)"b" },
		{ (unsigned char *)"D", (unsigned char *)"c" },
		{ (unsigned char *)"E", (unsigned char *)"d" },
		{ (unsigned char *)"F", (unsigned char *)"e" },
		{ (unsigned char *)"G", (unsigned char *)"f" },
		{ (unsigned char *)"H", (unsigned char *)"g" },
		{ (unsigned char *)"I", (unsigned char *)"h" },
		{ (unsigned char *)"J", (unsigned char *)"i" },
		{ (unsigned char *)"K", (unsigned char *)"j" },
		{ (unsigned char *)"L", (unsigned char *)"k" },
		{ (unsigned char *)"M", (unsigned char *)"l" },
		{ (unsigned char *)"N", (unsigned char *)"m" },
		{ (unsigned char *)"O", (unsigned char *)"n" },
		{ (unsigned char *)"P", (unsigned char *)"o" },
		{ (unsigned char *)"Q", (unsigned char *)"p" },
		{ (unsigned char *)"R", (unsigned char *)"q" },
		{ (unsigned char *)"S", (unsigned char *)"r" },
		{ (unsigned char *)"T", (unsigned char *)"s" },
		{ (unsigned char *)"U", (unsigned char *)"t" },
		{ (unsigned char *)"V", (unsigned char *)"u" },
		{ (unsigned char *)"W", (unsigned char *)"v" },
		{ (unsigned char *)"X", (unsigned char *)"w" },
		{ (unsigned char *)"Y", (unsigned char *)"x" },
		{ (unsigned char *)"Z", (unsigned char *)"y" },
		{ (unsigned char *)"[", (unsigned char *)"m" },
		{ (unsigned char *)"\\", (unsigned char *)"" },
		{ (unsigned char *)"]", (unsigned char *)"n" },
		{ (unsigned char *)"^", (unsigned char *)"O" },
		{ (unsigned char *)"_", (unsigned char *)"Q" },
		{ (unsigned char *)"`", (unsigned char *)"e" },
		{ (unsigned char *)"a", (unsigned char *)"" },
		{ (unsigned char *)"b", (unsigned char *)"" },
		{ (unsigned char *)"c", (unsigned char *)"" },
		{ (unsigned char *)"d", (unsigned char *)"" },
		{ (unsigned char *)"e", (unsigned char *)"" },
		{ (unsigned char *)"f", (unsigned char *)"" },
		{ (unsigned char *)"g", (unsigned char *)"" },
		{ (unsigned char *)"h", (unsigned char *)"" },
		{ (unsigned char *)"i", (unsigned char *)"" },
		{ (unsigned char *)"j", (unsigned char *)"" },
		{ (unsigned char *)"k", (unsigned char *)"" },
		{ (unsigned char *)"l", (unsigned char *)"" },
		{ (unsigned char *)"m", (unsigned char *)"" },
		{ (unsigned char *)"n", (unsigned char *)"" },
		{ (unsigned char *)"o", (unsigned char *)"" },
		{ (unsigned char *)"p", (unsigned char *)"" },
		{ (unsigned char *)"q", (unsigned char *)"" },
		{ (unsigned char *)"r", (unsigned char *)"" },
		{ (unsigned char *)"s", (unsigned char *)"" },
		{ (unsigned char *)"t", (unsigned char *)"" },
		{ (unsigned char *)"u", (unsigned char *)"" },
		{ (unsigned char *)"v", (unsigned char *)"" },
		{ (unsigned char *)"w", (unsigned char *)"" },
		{ (unsigned char *)"x", (unsigned char *)"" },
		{ (unsigned char *)"y", (unsigned char *)"" },
		{ (unsigned char *)"z", (unsigned char *)"" },
		{ (unsigned char *)"{", (unsigned char *)"o" },
		{ (unsigned char *)"|", (unsigned char *)"b" },
		{ (unsigned char *)"}", (unsigned char *)"p" },
		{ (unsigned char *)"~", (unsigned char *)"`" },
		{ (unsigned char *)"", (unsigned char *)"B" },
		{ (unsigned char *)"", (unsigned char *)"u" },
		{ (unsigned char *)"", (unsigned char *)"v" },
		{ (unsigned char *)"", (unsigned char *)"A" },
		{ (unsigned char *)"", (unsigned char *)"E" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"@" },
		{ (unsigned char *)"", (unsigned char *)"B" },
		{ (unsigned char *)"", (unsigned char *)"D" },
		{ (unsigned char *)"", (unsigned char *)"F" },
		{ (unsigned char *)"", (unsigned char *)"H" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"b" },
		{ (unsigned char *)"", (unsigned char *)"|" },
		{ (unsigned char *)"", (unsigned char *)"A" },
		{ (unsigned char *)"", (unsigned char *)"C" },
		{ (unsigned char *)"", (unsigned char *)"E" },
		{ (unsigned char *)"", (unsigned char *)"G" },
		{ (unsigned char *)"", (unsigned char *)"I" },
		{ (unsigned char *)"", (unsigned char *)"J" },
		{ (unsigned char *)"", (unsigned char *)"L" },
		{ (unsigned char *)"", (unsigned char *)"N" },
		{ (unsigned char *)"", (unsigned char *)"P" },
		{ (unsigned char *)"", (unsigned char *)"R" },
		{ (unsigned char *)"", (unsigned char *)"T" },
		{ (unsigned char *)"", (unsigned char *)"V" },
		{ (unsigned char *)"", (unsigned char *)"X" },
		{ (unsigned char *)"", (unsigned char *)"Z" },
		{ (unsigned char *)"", (unsigned char *)"\" },
		{ (unsigned char *)"", (unsigned char *)"^" },
		{ (unsigned char *)"", (unsigned char *)"`" },
		{ (unsigned char *)"", (unsigned char *)"c" },
		{ (unsigned char *)"", (unsigned char *)"e" },
		{ (unsigned char *)"", (unsigned char *)"g" },
		{ (unsigned char *)"", (unsigned char *)"i" },
		{ (unsigned char *)"", (unsigned char *)"j" },
		{ (unsigned char *)"", (unsigned char *)"k" },
		{ (unsigned char *)"", (unsigned char *)"l" },
		{ (unsigned char *)"", (unsigned char *)"m" },
		{ (unsigned char *)"", (unsigned char *)"n" },
		{ (unsigned char *)"", (unsigned char *)"q" },
		{ (unsigned char *)"", (unsigned char *)"t" },
		{ (unsigned char *)"", (unsigned char *)"w" },
		{ (unsigned char *)"", (unsigned char *)"z" },
		{ (unsigned char *)"", (unsigned char *)"}" },
		{ (unsigned char *)"", (unsigned char *)"~" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"" },
		{ (unsigned char *)"", (unsigned char *)"J" },
		{ (unsigned char *)"", (unsigned char *)"K" },
		{ (unsigned char *)"K", (unsigned char *)"JJ" },
		{ (unsigned char *)"M", (unsigned char *)"LJ" },
		{ (unsigned char *)"O", (unsigned char *)"NJ" },
		{ (unsigned char *)"Q", (unsigned char *)"PJ" },
		{ (unsigned char *)"S", (unsigned char *)"RJ" },
		{ (unsigned char *)"U", (unsigned char *)"TJ" },
		{ (unsigned char *)"W", (unsigned char *)"VJ" },
		{ (unsigned char *)"Y", (unsigned char *)"XJ" },
		{ (unsigned char *)"[", (unsigned char *)"ZJ" },
		{ (unsigned char *)"]", (unsigned char *)"\J" },
		{ (unsigned char *)"_", (unsigned char *)"^J" },
		{ (unsigned char *)"a", (unsigned char *)"`J" },
		{ (unsigned char *)"d", (unsigned char *)"cJ" },
		{ (unsigned char *)"f", (unsigned char *)"eJ" },
		{ (unsigned char *)"h", (unsigned char *)"gJ" },
		{ (unsigned char *)"o", (unsigned char *)"nJ" },
		{ (unsigned char *)"r", (unsigned char *)"qJ" },
		{ (unsigned char *)"u", (unsigned char *)"tJ" },
		{ (unsigned char *)"x", (unsigned char *)"wJ" },
		{ (unsigned char *)"{", (unsigned char *)"zJ" },
		{ (unsigned char *)"p", (unsigned char *)"nK" },
		{ (unsigned char *)"s", (unsigned char *)"qK" },
		{ (unsigned char *)"v", (unsigned char *)"tK" },
		{ (unsigned char *)"y", (unsigned char *)"wK" },
		{ (unsigned char *)"|", (unsigned char *)"zK" },
		{ (unsigned char *)"", (unsigned char *)"EJ" },
		{ (unsigned char *)NULL, NULL }
	}, *c;
	unsigned char *p;
	const unsigned char *q;

	p = out;
	for(q = in; *q != 0; q += (isKanji(*q) ? 2: 1)) {
		for(c = conv; c->src != NULL; c++)
			if(memcmp(q, c->src, strlen((char *)c->src)) == 0)
				break;

		if (c->src == NULL) {
			memcpy(p, q, isKanji(*q) ? 2: 1);
			p += (isKanji(*q) ? 2: 1);
		} else {
			memcpy(p, c->dst, strlen((char *)c->dst));
			p += strlen((char *)c->dst);
		}
	}
	*p = 0;

	return strlen((char *)out);
}

/*
	5TQt@Cǂݍ
*/
static int read5tq(unsigned char **gotaku, int **n, const char *path)
{
	int fd, i, j, size;
	unsigned char category[16 + 1], q[116 + 1], a1[28 + 1], a2[28 + 1], a3[28 + 1], a4[28 + 1], a5[28 + 1], *p;

	*gotaku = p = malloc(65536);
	*n = malloc(sizeof(int) * 8);

	if((fd = open(path, O_RDONLY | O_BINARY)) < 0)
		goto fail;

	/* wb_ */
	for(i = 0; i < 8; i++) {
		if(read5tqHead(fd, category, &((*n)[i])) < 0)
			goto fail;
		p += replaceChar(p, category) + 1;
	}

	/* f[^ */
	for(i = 0; i < 8; i++)
		for(j = 0; j < (*n)[i]; j++) {
			if(read5tqData(fd, q, a1, a2, a3, a4, a5) < 0)
				goto fail;
			p += replaceChar(p, q) + 1;
			p += replaceChar(p, a1) + 1;
			p += replaceChar(p, a2) + 1;
			p += replaceChar(p, a3) + 1;
			p += replaceChar(p, a4) + 1;
			p += replaceChar(p, a5) + 1;
		}
	*p++ = 0;

	close(fd);

	size = (int )(p - *gotaku);
	*gotaku = realloc(*gotaku, size);
	return size;

fail:;
	free(*n);
	free(*gotaku);
	if(fd >= 0)
		close(fd);
	return -1;
}

/*
	R[h\ɉ
*/
static struct CodeTable *addCode(struct CodeTable *code_table, const unsigned char *buf)
{
	int i;
	const unsigned char *p;

	for(p = buf; *p != 0; p += (isKanji(*p) ? 2: 1)) {
		if(isKanji(*p)) {
			for(i = 0; code_table[i].n != 0; i++)
				if(memcmp(p, code_table[i].ch, 2) == 0)
					break;
		} else {
			for(i = 0; code_table[i].n != 0; i++)
				if(*p == code_table[i].ch[0])
					break;
		}
		
		if(code_table[i].n == 0) {
			code_table = realloc(code_table, sizeof(struct CodeTable) * (i + 2));
			code_table[i + 1].n = 0;
			if(isKanji(*p))
				memcpy(code_table[i].ch, p, 2);
			else {
				code_table[i].ch[0] = *p;
				code_table[i].ch[1] = 0;
			}
		}
		code_table[i].n++;
	}

	return code_table;
}

/*
	̏or
*/
static int cmpCode(const void *a, const void *b)
{
	const struct CodeTable *c1 = a, *c2 = b;

	return c2->n - c1->n;
}

/*
	R[h\
*/
static int compCode(struct CodeTable *code_table)
{
	struct CodeTable *c;

	for(c = code_table; c->n != 0; c++)
		;

	qsort(code_table, (int )(c - code_table), sizeof(code_table[0]), cmpCode);

	return (int )(c - code_table);
}

/*
	5TQt@CR[h\
*/
static struct CodeTable *makeCodeTable(const unsigned char *gotaku)
{
	struct CodeTable *code_table;
	const unsigned char *p;

	code_table = malloc(sizeof(struct CodeTable));
	code_table[0].n = 0;

	for(p = gotaku; *p != 0; p += strlen((char *)p) + 1)
		code_table = addCode(code_table, p);
	compCode(code_table);

	return code_table;
}

/*
	R[h𓾂
*/
static unsigned short findCode(const struct CodeTable *code_table, const unsigned char *ch)
{
	const struct CodeTable *c;

	for(c = code_table; c->n != 0; c++)
		if(memcmp(ch, c->ch, 2) == 0)
			return (unsigned short )(c - code_table);
	/*
	printf("findCode ERROR! [%2.2s]\n", ch);
	*/
	return 0;
}

/*
	̐߂
*/
static int getLines16(const unsigned short *code16)
{
	int lines = 0;
	const unsigned short *p;

	for(p = code16, lines = 0; *p != 0; p += *p + 1, lines++)
		;
	return lines;
}

/*
	R[h̐߂
*/
static int getCodes16(const unsigned short *code16)
{
	int codes = 0;
	const unsigned short *p, *q;

	for(p = code16; *p != 0; p += *p + 1)
		for(q = p; q < p + *p + 1; q++)
			if(codes <= *q)
				codes = *q + 1;
	return codes;
}

/*
	5TQ16bitR[hɂ
*/
static unsigned short *encode16(const struct CodeTable *code_table, const unsigned char *gotaku)
{
	int size, n;
	unsigned short *code16, *p;
	const unsigned char *q, *r;

	code16 = p = malloc(sizeof(unsigned short) * 65536);

	for(q = gotaku; *q != 0; q += strlen((char *)q) + 1) {
		for(r = q, n = 0; *r != 0; r += (isKanji(*r) ? 2: 1), n++)
			;
		*p++ = n;
		for(r = q; *r != 0; r += (isKanji(*r) ? 2: 1))
			*p++ = findCode(code_table, r);
	}
	*p++ = 0;

	size = (int )(p - code16);
	/*code16 = realloc(code16, size * sizeof(unsigned short));*/
	return code16;
}

/*
	R[h8bitƂ̃TCY𓾂
*/
static int getCode8Size(const unsigned short *code16, int code16_size, unsigned char c2)
{
	int size = 0;
	const unsigned short *p;

	for(p = code16; p < code16 + code16_size; p++)
		size += (*p < (unsigned short )c2 ? 1: 2);
	return size;
}

/*
	16bitR[hk
*/
static int pack(unsigned short *code16, const unsigned short *dic, int dic_len, unsigned char c2, unsigned short new_code)
{
	int size, dic_size, count = 0;
	unsigned short tmp[128], *line, *p;

	/* ^ꂽƈv񐔂߂ */
	for(line = code16; *line != 0; line += *line + 1) {
		for(p = line + 1; p <= (line + 1) + *line - dic_len; p++) {
			if(memcmp(p, dic, dic_len * sizeof(unsigned short)) == 0) {
				p += dic_len - 1;
				count++;
			}
		}
	}

	/* ^ꂽɉƂȂ邩? */
	dic_size = getCode8Size(dic, dic_len, c2);
	if(dic_size * count <= 2 * count + 1 + dic_size)
		return FALSE;

	size = ((int )(line - code16) + 1) * sizeof(unsigned short);
	memcpy(tmp, dic, dic_len * sizeof(unsigned short));

	/* R[hɒu */
	for(line = code16; *line != 0; line += *line + 1) {
		for(p = line + 1; p <= (line + 1) + *line - dic_len; p++) {
			if(memcmp(p, tmp, dic_len * sizeof(unsigned short)) == 0) {
				*p = new_code;
				*line -= dic_len - 1;
				size -= (dic_len - 1) * sizeof(unsigned short);
				memmove(p + 1, p + dic_len, size);
			} else
				size -= sizeof(unsigned short);
		}
	}

	/* ɉ */
	*line++ = dic_len;
	memcpy(line, tmp, dic_len * sizeof(unsigned short));
	line += dic_len;
	*line = 0;

	/*
	printf("HIT code=%d, len=%d, count=%d\n", new_code, dic_len, count);
	*/
	return TRUE;
}

/*
	2R[h߂
*/
unsigned char getC2Code(int codes)
{
	return (0x80 * 0x80 - codes) / (0x80 - 1);
}

/*
	16bitR[hk
*/
static void pack16(unsigned short *code16)
{
	int dic_len, codes = getCodes16(code16);
	unsigned short *line, *dic, new_code = (unsigned short )codes;
	unsigned char c2 = getC2Code(codes);

	for(dic_len = 116; dic_len >= 2; dic_len--) {
		fprintf(stderr, "[%03d]\r", dic_len - 2);
		fflush(stderr);

		for(line = code16; *line != 0; line += *line + 1)
			for(dic = line + 1; dic <= (line + 1) + *line - dic_len; dic++)
				if(pack(code16, dic, dic_len, c2, new_code))
					new_code++;
	}
}

/*
	16bitR[h18bit
*/
static int encode8_1(unsigned char *out, unsigned char c2, const unsigned short in)
{
	unsigned short code;

	if(in < (unsigned short )c2) {
		out[0] = (unsigned char )in;
		return 1;
	}

	code = in - (unsigned short )c2;

	out[0] = (unsigned short )c2 + code / 0x80;
	out[1] = code % 0x80;
	return 2;
}

/*
	16bitR[h8bit
*/
static unsigned char *encode8(int *len, const unsigned short *code16, int dics, unsigned short dic_code, const int *n)
{
	int i, j, codes, lines, sum = 0;
	const unsigned short *p, *q;
	unsigned char c2, *code8, *r;

	codes = getCodes16(code16);
	lines = getLines16(code16);
	code8 = r = malloc(65536);

	/* 2f[^JnR[h𓾂 */
	c2 = getC2Code(codes);

	/* 2f[^JnR[h */
	r += encode8_1(r, c2, c2 - 1);

	/* R[h */
	r += encode8_1(r, c2, dic_code);

	/*  */
	r += encode8_1(r, c2, dics);

	/* 8bit */
	for(j = 1; j <= dics; j++) {
		for(i = 0, p = code16; i < lines - j; i++, p += *p + 1)
			;
		for(q = p; q < p + *p + 1; q++)
			r += encode8_1(r, c2, *q < dic_code ? *q: dic_code + (codes - *q - 1));
	}

	/* wb_8bit */
	for(i = 0, p = code16; i < 8; i++, p += *p + 1) {
		for(q = p; q < p + *p + 1; q++)
			r += encode8_1(r, c2, *q < dic_code ? *q: dic_code + (codes - *q - 1));
		r += encode8_1(r, c2, sum);
		r += encode8_1(r, c2, n[i]);
		sum += n[i];
	}

	/* {8bit */
	for(; i < lines - dics; i++, p += *p + 1)
		for(q = p; q < p + *p + 1; q++)
			r += encode8_1(r, c2, *q < dic_code ? *q: dic_code + (codes - *q - 1));

	*len = (int )(r - code8);
	return code8;
}

/*
	8bitR[h7bitR[hɂď
*/
static int write7(unsigned char *out, const unsigned char *in, int n)
{
	int out_seg, out_off, in_seg, in_off;
	unsigned char out_mask, in_mask;

	for(in_seg = 0, out_seg = 0; in_seg < n; in_seg += 8, out_seg += 7) {
		for(in_off = 0, out_mask = 1; in_off < 8; in_off++, out_mask <<= 1) {
			for(out_off = 6, in_mask = 1; out_off >= 0; out_off--, in_mask <<= 1) {
				if(in[in_seg + in_off] & in_mask)
					out[out_seg + out_off] |= out_mask;
				else
					out[out_seg + out_off] &= ~out_mask;
			}
		}
	}

	return out_seg;
}

/*
	SJISJISɕϊ
*/
static unsigned char *sjis2jis(unsigned char *jis, const unsigned char *sjis)
{
	jis[0] = sjis[0];
	jis[1] = sjis[1];

	if(!(jis[0] & 0x80))
		return jis;

	if(jis[0] >= 0xe0)
		jis[0] -= 0x40;
	if(jis[1] >= 0x80)
		jis[1]--;
	if(jis[1] >= 0x9e) {
		jis[0] = (jis[0] - 0x70) << 1;
		jis[1] -= 0x7d;
	} else {
		jis[0] = ((jis[0] - 0x70) << 1) - 1;
		jis[1] -= 0x1f;
	}
	return jis;
}

/*
	tHgf[^
*/
static unsigned char *writeFont(unsigned char *font, const unsigned char *pat, int width)
{
	int i;

	memset(font, 0, 7);
	for(i = 0; i < width; i++)
		if(pat[i] != 0) {
			memcpy(font, &pat[i], width - i);
			return font;
		}
	return font;
}

/*
	tHgf[^쐬
*/
static int encodeFont(unsigned char *out, const struct CodeTable *code_table)
{
	const struct CodeTable *c;
	int off;
	unsigned char jis[2], code8[65535], *p = code8;

	for(c = code_table; c->n != 0; c++) {
		sjis2jis(jis, c->ch);
		off = ((int )jis[0] - 0x21) * 94 + ((int )jis[1] - 0x21);
		
		if(jis[0] == 0x21) {
			if(jis[1] == 0x21) /*   */
				writeFont(p, font4x8[0x20], 3);
			else if(jis[1] == 0x22) /* A */
				writeFont(p, font4x8[0xa4], 3);
			else if(jis[1] == 0x23) /* B */
				writeFont(p, font4x8[0xa1], 3);
			else if(jis[1] == 0x24) /* , */
				writeFont(p, font4x8[0x2c], 3);
			else if(jis[1] == 0x25) /* . */
				writeFont(p, font4x8[0x2e], 3);
			else if(jis[1] == 0x26) /* E */
				writeFont(p, font4x8[0xa5], 3);
			else if(jis[1] == 0x27) /* : */
				writeFont(p, font4x8[0x3a], 3);
			else if(jis[1] == 0x28) /* ; */
				writeFont(p, font4x8[0x3b], 3);
			else if(jis[1] == 0x29) /* ? */
				writeFont(p, font4x8[0x3f], 3);
			else if(jis[1] == 0x2a) /* ! */
				writeFont(p, font4x8[0x21], 3);
			else if(jis[1] == 0x2b) /* J */
				writeFont(p, font4x8[0xde], 3);
			else if(jis[1] == 0x2c) /* K */
				writeFont(p, font4x8[0xdf], 3);
			else if(jis[1] == 0x30) /* ^ */
				writeFont(p, font4x8[0x5e], 3);
			else if(jis[1] == 0x31) /* P */
				writeFont(p, font4x8[0x7e], 3);
			else if(jis[1] == 0x32) /* Q */
				writeFont(p, font4x8[0x5f], 3);
			else if(jis[1] == 0x3c) /* [ */
				writeFont(p, font4x8[0xb0], 3);
			else if(jis[1] == 0x3f) /* / */
				writeFont(p, font4x8[0x2f], 3);
			else if(jis[1] == 0x43) /* | */
				writeFont(p, font4x8[0x7c], 3);
			else if(jis[1] == 0x46) /* ` */
				writeFont(p, font4x8[0x60], 3);
			else if(jis[1] == 0x47) /* ' */
				writeFont(p, font4x8[0x27], 3);
			else if(jis[1] == 0x4a) /* ( */
				writeFont(p, font4x8[0x28], 3);
			else if(jis[1] == 0x4b) /* ) */
				writeFont(p, font4x8[0x29], 3);
			else if(jis[1] == 0x4e) /* [ */
				writeFont(p, font4x8[0x5b], 3);
			else if(jis[1] == 0x4f) /* ] */
				writeFont(p, font4x8[0x5d], 3);
			else if(jis[1] == 0x50) /* { */
				writeFont(p, font4x8[0x7b], 3);
			else if(jis[1] == 0x51) /* } */
				writeFont(p, font4x8[0x7d], 3);
			else if(jis[1] == 0x56) /* u */
				writeFont(p, font4x8[0xa2], 3);
			else if(jis[1] == 0x57) /* v */
				writeFont(p, font4x8[0xa3], 3);
			else if(jis[1] == 0x5c) /* + */
				writeFont(p, font4x8[0x2b], 3);
			else if(jis[1] == 0x5d) /* - */
				writeFont(p, font4x8[0x2d], 3);
			else if(jis[1] == 0x61) /* = */
				writeFont(p, font4x8[0x3d], 3);
			else if(jis[1] == 0x63) /* < */
				writeFont(p, font4x8[0x3c], 3);
			else if(jis[1] == 0x64) /* > */
				writeFont(p, font4x8[0x3e], 3);
			else if(jis[1] == 0x6f) /* \ */
				writeFont(p, font4x8[0x5c], 3);
			else if(jis[1] == 0x70) /* $ */
				writeFont(p, font4x8[0x24], 3);
			else if(jis[1] == 0x73) /* % */
				writeFont(p, font4x8[0x25], 3);
			else if(jis[1] == 0x74) /* # */
				writeFont(p, font4x8[0x23], 3);
			else if(jis[1] == 0x75) /* & */
				writeFont(p, font4x8[0x26], 3);
			else if(jis[1] == 0x76) /* * */
				writeFont(p, font4x8[0x2a], 3);
			else if(jis[1] == 0x77) /* @ */
				writeFont(p, font4x8[0x40], 3);
			else
				writeFont(p, fontK6x8[off], 5);
		} else  if(jis[0] == 0x23) { /* p */
			writeFont(p, font4x8[jis[1]], 3);
		} else if(jis[0] == 0x25) { /* J^Ji */
			if(jis[1] == 0x21) /* @ */
				writeFont(p, font4x8[0xa7], 3);
			else if(jis[1] == 0x22) /* A */
				writeFont(p, font4x8[0xb1], 3);
			else if(jis[1] == 0x23) /* B */
				writeFont(p, font4x8[0xa8], 3);
			else if(jis[1] == 0x24) /* C */
				writeFont(p, font4x8[0xb2], 3);
			else if(jis[1] == 0x25) /* D */
				writeFont(p, font4x8[0xa9], 3);
			else if(jis[1] == 0x26) /* E */
				writeFont(p, font4x8[0xb3], 3);
			else if(jis[1] == 0x27) /* F */
				writeFont(p, font4x8[0xaa], 3);
			else if(jis[1] == 0x28) /* G */
				writeFont(p, font4x8[0xb4], 3);
			else if(jis[1] == 0x29) /* H */
				writeFont(p, font4x8[0xab], 3);
			else if(jis[1] == 0x2a) /* I */
				writeFont(p, font4x8[0xb5], 3);
			else if(jis[1] == 0x2b) /* J */
				writeFont(p, font4x8[0xb6], 3);
			else if(jis[1] == 0x2c) /* K */
				writeFont(p, font4x8[0xb6], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x2d) /* L */
				writeFont(p, font4x8[0xb7], 3);
			else if(jis[1] == 0x2e) /* M */
				writeFont(p, font4x8[0xb7], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x2f) /* N */
				writeFont(p, font4x8[0xb8], 3);
			else if(jis[1] == 0x30) /* O */
				writeFont(p, font4x8[0xb8], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x31) /* P */
				writeFont(p, font4x8[0xb9], 3);
			else if(jis[1] == 0x32) /* Q */
				writeFont(p, font4x8[0xb9], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x33) /* R */
				writeFont(p, font4x8[0xba], 3);
			else if(jis[1] == 0x34) /* S */
				writeFont(p, font4x8[0xba], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x35) /* T */
				writeFont(p, font4x8[0xbb], 3);
			else if(jis[1] == 0x36) /* U */
				writeFont(p, font4x8[0xbb], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x37) /* V */
				writeFont(p, font4x8[0xbc], 3);
			else if(jis[1] == 0x38) /* W */
				writeFont(p, font4x8[0xbc], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x39) /* X */
				writeFont(p, font4x8[0xbd], 3);
			else if(jis[1] == 0x3a) /* Y */
				writeFont(p, font4x8[0xbd], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x3b) /* Z */
				writeFont(p, font4x8[0xbe], 3);
			else if(jis[1] == 0x3c) /* [ */
				writeFont(p, font4x8[0xbe], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x3d) /* \ */
				writeFont(p, font4x8[0xbf], 3);
			else if(jis[1] == 0x3e) /* ] */
				writeFont(p, font4x8[0xbf], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x3f) /* ^ */
				writeFont(p, font4x8[0xc0], 3);
			else if(jis[1] == 0x40) /* _ */
				writeFont(p, font4x8[0xc0], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x41) /* ` */
				writeFont(p, font4x8[0xc1], 3);
			else if(jis[1] == 0x42) /* a */
				writeFont(p, font4x8[0xc1], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x43) /* b */
				writeFont(p, font4x8[0xaf], 3);
			else if(jis[1] == 0x44) /* c */
				writeFont(p, font4x8[0xc2], 3);
			else if(jis[1] == 0x45) /* d */
				writeFont(p, font4x8[0xc2], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x46) /* e */
				writeFont(p, font4x8[0xc3], 3);
			else if(jis[1] == 0x47) /* f */
				writeFont(p, font4x8[0xc3], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x48) /* g */
				writeFont(p, font4x8[0xc4], 3);
			else if(jis[1] == 0x49) /* h */
				writeFont(p, font4x8[0xc4], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x4a) /* i */
				writeFont(p, font4x8[0xc5], 3);
			else if(jis[1] == 0x4b) /* j */
				writeFont(p, font4x8[0xc6], 3);
			else if(jis[1] == 0x4c) /* k */
				writeFont(p, font4x8[0xc7], 3);
			else if(jis[1] == 0x4d) /* l */
				writeFont(p, font4x8[0xc8], 3);
			else if(jis[1] == 0x4e) /* m */
				writeFont(p, font4x8[0xc9], 3);
			else if(jis[1] == 0x4f) /* n */
				writeFont(p, font4x8[0xca], 3);
			else if(jis[1] == 0x50) /* o */
				writeFont(p, font4x8[0xca], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x51) /* p */
				writeFont(p, font4x8[0xca], 3), writeFont(p + 4, font4x8[0xdf], 3);
			else if(jis[1] == 0x52) /* q */
				writeFont(p, font4x8[0xcb], 3);
			else if(jis[1] == 0x53) /* r */
				writeFont(p, font4x8[0xcb], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x54) /* s */
				writeFont(p, font4x8[0xcb], 3), writeFont(p + 4, font4x8[0xdf], 3);
			else if(jis[1] == 0x55) /* t */
				writeFont(p, font4x8[0xcc], 3);
			else if(jis[1] == 0x56) /* u */
				writeFont(p, font4x8[0xcc], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x57) /* v */
				writeFont(p, font4x8[0xcc], 3), writeFont(p + 4, font4x8[0xdf], 3);
			else if(jis[1] == 0x58) /* w */
				writeFont(p, font4x8[0xcd], 3);
			else if(jis[1] == 0x59) /* x */
				writeFont(p, font4x8[0xcd], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x5a) /* y */
				writeFont(p, font4x8[0xcd], 3), writeFont(p + 4, font4x8[0xdf], 3);
			else if(jis[1] == 0x5b) /* z */
				writeFont(p, font4x8[0xce], 3);
			else if(jis[1] == 0x5c) /* { */
				writeFont(p, font4x8[0xce], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x5d) /* | */
				writeFont(p, font4x8[0xce], 3), writeFont(p + 4, font4x8[0xdf], 3);
			else if(jis[1] == 0x5e) /* } */
				writeFont(p, font4x8[0xcf], 3);
			else if(jis[1] == 0x5f) /* ~ */
				writeFont(p, font4x8[0xd0], 3);
			else if(jis[1] == 0x60) /*  */
				writeFont(p, font4x8[0xd1], 3);
			else if(jis[1] == 0x61) /*  */
				writeFont(p, font4x8[0xd2], 3);
			else if(jis[1] == 0x62) /*  */
				writeFont(p, font4x8[0xd3], 3);
			else if(jis[1] == 0x63) /*  */
				writeFont(p, font4x8[0xac], 3);
			else if(jis[1] == 0x64) /*  */
				writeFont(p, font4x8[0xd4], 3);
			else if(jis[1] == 0x65) /*  */
				writeFont(p, font4x8[0xad], 3);
			else if(jis[1] == 0x66) /*  */
				writeFont(p, font4x8[0xd5], 3);
			else if(jis[1] == 0x67) /*  */
				writeFont(p, font4x8[0xae], 3);
			else if(jis[1] == 0x68) /*  */
				writeFont(p, font4x8[0xd6], 3);
			else if(jis[1] == 0x69) /*  */
				writeFont(p, font4x8[0xd7], 3);
			else if(jis[1] == 0x6a) /*  */
				writeFont(p, font4x8[0xd8], 3);
			else if(jis[1] == 0x6b) /*  */
				writeFont(p, font4x8[0xd9], 3);
			else if(jis[1] == 0x6c) /*  */
				writeFont(p, font4x8[0xda], 3);
			else if(jis[1] == 0x6d) /*  */
				writeFont(p, font4x8[0xdb], 3);
			else if(jis[1] == 0x6e) /*  */
				writeFont(p, font4x8[0xe0], 5);
			else if(jis[1] == 0x6f) /*  */
				writeFont(p, font4x8[0xdc], 3);
			else if(jis[1] == 0x70) /*  */
				writeFont(p, font4x8[0xe3], 5);
			else if(jis[1] == 0x71) /*  */
				writeFont(p, font4x8[0xe4], 5);
			else if(jis[1] == 0x72) /*  */
				writeFont(p, font4x8[0xa6], 3);
			else if(jis[1] == 0x73) /*  */
				writeFont(p, font4x8[0xdd], 3);
			else if(jis[1] == 0x74) /*  */
				writeFont(p, font4x8[0xb3], 3), writeFont(p + 4, font4x8[0xde], 3);
			else if(jis[1] == 0x75) /*  */
				writeFont(p, font4x8[0xe1], 5);
			else if(jis[1] == 0x76) /*  */
				writeFont(p, font4x8[0xe2], 5);
			else
				writeFont(p, fontK6x8[off], 5);
		} else if(0x21 <= jis[0] && jis[0] <= 0x28) { /* 񊿎 */
			writeFont(p, fontK6x8[off], 5);
		} else { /*  */
			writeFont(p, fontMisaki[off], 7);
		}
		p += 7;
	}

	return write7(out, code8, (int )(p - code8));
}

/*
	ƃtHg7bit
*/
static unsigned char *encode7(int *code7_size, int base, const struct CodeTable *code_table, unsigned char *code8, int code8_size)
{
	int q_size, f_size;
	unsigned char *code7;

	code7 = malloc(65535);

	/* nCXRA */
	code7[0] = 0;
	code7[1] = 0;

	/* f[^ */
	q_size = write7(&code7[4], code8, code8_size);

	/* tHg */
	f_size = encodeFont(&code7[4 + q_size], code_table);

	/* tHg̃AhX */
	code7[2] = (base + 4 + q_size) % 0x100;
	code7[3] = (base + 4 + q_size) / 0x100;

	*code7_size = 4 + q_size + f_size;
	return code7;
}

/*
	t@Cgq菜𓾂
*/
static char *getBaseFileName (char *base, const char *file)
{
	char *p;

	strcpy(base, file);
	for(p = base + strlen(base) - 1; p > base && *p != '.'; p--)
		*p = 0;
	if(*p == '.')
		*p = 0;
	else
		strcpy(base, file);

#if !defined(WIN32) && !defined(_WIN32) && !defined(__WIN32)
	for(p = base + strlen(base) - 1; p > base && *p != '/'; p--)
		;
	if(*p == '/')
		memmove(base, p + 1, strlen(p + 1) + 1);
#endif

	return base;
}

int main(int argc, char *argv[])
{
	FILE *fp;
	struct CodeTable *code_table;
	int *n, i, lines, codes, dics, code8_size, code7_size;
	const int base = 0x5d0;
	unsigned short dic_code, *code16;
	char file[PATH_MAX], path[PATH_MAX], logfile[PATH_MAX];
	unsigned char *gotaku, *code8, *code7;

	if(argc == 1) {
		fprintf(stderr, "USAGE: %s [5TQ file] ...\n", argv[0]);
		return 0;
	}

	sprintf(logfile, "%s.log", getBaseFileName(file, argv[1]));
	unlink(logfile);

	for(i = 1; i < argc; i++) {
		gotaku = NULL;
		code16 = NULL;
		code8 = NULL;
		code7 = NULL;

		/* 5TQt@Cǂݍ */
		if(read5tq(&gotaku, &n, argv[i]) < 0) {
			fprintf(stderr, "Cannot open %s\n", argv[i]);
			goto fail;
		}
		/*
		{
			char *p;

			for(p = gotaku; *p != 0; p += strlen(p) + 1)
				printf("%s\n", p);
		}
		*/

		/* R[h\𐶐 */
		if((code_table = makeCodeTable(gotaku)) == NULL) {
			fprintf(stderr, "Cannot open %s\n", argv[i]);
			goto fail;
		}
		/*
		{
			int i, sum = 0;
			struct CodeTable *c;

			for(c = code_table; c->n != 0; c++) {
				sum += c->n;
				printf("0x%02x\t%d\t%d\t%2.2s\t", (int )(c - code_table), c->n, sum, c->ch);
				for(i = 0; i < 40 * c->n / code_table[0].n; i++)
					printf("*");
				printf("\n");
			}
		}
		*/

		/* R[hɒu */
		code16 = encode16(code_table, gotaku);
		lines = getLines16(code16);
		codes = getCodes16(code16);
		/*
		{
			unsigned short *p;
	
			for(p = code16; *p != 0; p += *p + 1)
				printf("%d\n", *p);
		}
		*/

		/* R[hk */
		pack16(code16);
		dics = getLines16(code16) - lines;
		dic_code = codes;
		/*
		{
			struct CodeTable *c;
			int i, n, line = 0;
			unsigned short *p, *q;
	
			for(c = code_table, n = 0; c->n != 0; c++, n++)
				;
	
			printf("dic_code=0x%04x\n", dic_code);
			printf("dic_lines=0x%04x\n", dics);
			for(i = 0, p = code16; i < lines; i++, p += *p + 1)
				;
			for(; *p != 0; p += *p + 1) {
				printf("%d,%d,", line++, *p);
				for(q = p + 1; q < p + *p + 1; q++) {
					if(*q >= n)
						printf("[%d]", *q - dic_code);
					else
						printf("%2.2s", code_table[*q].ch);
				}
				printf("\n");
			}
		
			for(i = 0, p = code16; i < lines; i++, p += *p + 1) {
				printf("%d,%d,", line++, *p);
				for(q = p + 1; q < p + *p + 1; q++) {
					if(*q >= n)
						printf("[%d]", *q - dic_code);
					else
						printf("%2.2s", code_table[*q].ch);
				}
				printf("\n");
			}
		}
		*/

		/* R[h8bit */
		code8 = encode8(&code8_size, code16, dics, dic_code, n);
		/*
		{
			unsigned char *p;

			for(p = code8; p < code8 + code8_size; p++)
				printf("0x%02x\n", *p);
			printf("\n");
		}
		*/

		/* R[h7bit */
		code7 = encode7(&code7_size, base, code_table, code8, code8_size);

		/* (sԍȂ) */
		getBaseFileName(file, argv[i]);
		sprintf(path, "%s.ihx", file);
		if(writeHexAbs(path, code7, base, code7_size, FALSE) < 0) {
			fprintf(stderr, "Cannot write %s\n", path);
			goto fail;
		}
	
		/* (sԍ) */
		sprintf(path, "%s.txt", file);
		if(writeHexAbs(path, code7, base, code7_size, TRUE) < 0) {
			fprintf(stderr, "Cannot write %s\n", path);
			goto fail;
		}

		/* Oo͂ */
		if((fp = fopen(logfile, "a" FOPEN_TEXT)) != NULL) {
			fprintf(fp, "%s:%04X-%04X\n", argv[i], base, base + code7_size - 1);
			printf("%s:%04X-%04X\n", argv[i], base, base + code7_size - 1);
			fclose(fp);
		}

fail:;
		free(gotaku);
		free(encode16);
		free(encode8);
		free(encode7);
	}
	return 0;
}

/*
	{\tgEFA5tq2hex, ^Ȃނ6*8hbg{rbg}bvtH
	g k6x8 period 2, tHg, ꕔǉEύXĎgpĂ܂.

	--- (k6x8, tHg̃CZX ) --------------------------------------------
	These fonts are free softwares.
	Unlimited permission is granted to use, copy, and distribute it, with or without modification, either commercially and noncommercially.
	THESE FONTS ARE PROVIDED "AS IS" WITHOUT WARRANTY.

	̃tHg̓t[iRȁj\tgEGAłB
	ς̗LɊւ炸A܂ƓIȗpłĂARɂpAAĔzz邱Ƃł܂ASĖۏ؂ƂĂ܂B

	Copyright(C) 2000-2007 Num Kadoma
	--- (k6x8, tHg̃CZX ܂) --------------------------------------------
*/

/*
	5tq2hexBSDX^CCZXɏ].

	Copyright 2011 maruhiro
	All rights reserved. 

	Redistribution and use in source and binary forms, 
	with or without modification, are permitted provided that 
	the following conditions are met: 

	 1. Redistributions of source code must retain the above copyright notice, 
	    this list of conditions and the following disclaimer. 

	 2. Redistributions in binary form must reproduce the above copyright notice, 
	    this list of conditions and the following disclaimer in the documentation 
	    and/or other materials provided with the distribution. 

	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
	THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* eof */
